#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import fcntl
import logging
import os
import sys
import time
from copy import copy
from functools import wraps
from itertools import islice
from logging import Formatter
from typing import Iterator, Any, Union, List, Callable

import numpy as np

try:
    from memory_profiler import memory_usage
except ImportError:
    memory_usage = lambda: [0]

from ruamel.yaml import YAML
from termcolor import colored

__all__ = ['get_sys_info', 'get_optimal_sample_size',
           'get_perm', 'time_profile', 'set_logger',
           'batch_iterator', 'batching', 'yaml',
           'profile_logger', 'load_contrib_module',
           'parse_arg', 'profiling', 'FileLock', 'train_required', 'get_first_available_gpu']


def get_first_available_gpu():
    try:
        import GPUtil
        r = GPUtil.getAvailable(order='random',
                                maxMemory=0.5,
                                maxLoad=0.5,
                                limit=1)
        if r:
            return r[0]
        raise ValueError
    except ImportError:
        return -1
    except ValueError:
        return -1


class FileLock:
    """
    Implements the Posix based file locking (Linux, Ubuntu, MacOS, etc.)
    """

    def __init__(self, lock_file: str = "LOCK"):
        self._lock_file = lock_file
        self._lock_file_fd = None

    @property
    def is_locked(self):
        return self._lock_file_fd is not None

    def acquire(self):
        open_mode = os.O_RDWR | os.O_CREAT | os.O_TRUNC
        fd = os.open(self._lock_file, open_mode)

        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self._lock_file_fd = fd
            return fd
        except (IOError, OSError):
            os.close(fd)
        return None

    def release(self):
        if self.is_locked:
            fd = self._lock_file_fd
            self._lock_file_fd = None
            fcntl.flock(fd, fcntl.LOCK_UN)
            os.close(fd)


def get_sys_info():
    from psutil import virtual_memory
    mem = virtual_memory()
    # get available memory in (M)
    avai = mem.available / 1e6

    def timer(x, y):
        stime = time.time()
        np.matmul(x, y)
        return time.time() - stime

    x = np.random.random([1000, 1000])
    y = np.random.random([1000, 1000])
    unit_time = timer(x, y)
    return avai, unit_time


def touch_dir(base_dir: str) -> None:
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)


def ralloc_estimator(n_lines, num_dim, unit_time, max_mem, max_time=60):
    est_time = num_dim * num_dim * n_lines / 1e9 * unit_time * 2
    est_mem = 60 + 30 * (n_lines * num_dim / 768 / 10000)
    if (est_time < max_time) and (est_mem < max_mem * 0.5):
        return n_lines
    return ralloc_estimator(int(n_lines * 0.9), num_dim, unit_time, max_mem, max_time)


def get_optimal_sample_size(x):
    max_mem, unit_time = get_sys_info()
    num_samples, num_dim = x.shape
    return ralloc_estimator(num_samples, num_dim, unit_time, max_mem, 30)


def get_perm(L, m):
    n = int(len(L) / m)
    avg = sum(L) / len(L) * m
    LR = sorted(enumerate(L), key=lambda x: -x[1])
    L = np.reshape([i[1] for i in LR], [m, n])
    R = np.reshape([i[0] for i in LR], [m, n])
    F = np.zeros([m, n])

    reranked = []
    for _ in range(n):
        ind = 0
        for i in range(m):
            if i % 2 == 0:
                start, direction = 0, 1
            else:
                start, direction = n - 1, -1
            while F[i, start] == 1:
                start += direction
            if (ind + L[i, start] < avg) or (direction == 1):
                ind += L[i, start]
                F[i, start] = 1
                reranked.append(R[i, start])
            else:
                start, direction = n - 1, -1
                while F[i, start] == 1:
                    start += direction
                ind += L[i, start]
                F[i, start] = 1
                reranked.append(R[i, start])

    return reranked


def time_profile(func):
    @wraps(func)
    def arg_wrapper(*args, **kwargs):
        if os.environ.get('GNES_PROFILING', False):
            start_t = time.perf_counter()
            if os.environ.get('GNES_PROFILING_MEM', False):
                start_mem = memory_usage()[0]
            r = func(*args, **kwargs)
            elapsed = time.perf_counter() - start_t
            if os.environ.get('GNES_PROFILING_MEM', False):
                end_mem = memory_usage()[0]
            # level_prefix = ''.join('-' for v in inspect.stack() if v and v.index is not None and v.index >= 0)
            level_prefix = ''
            if os.environ.get('GNES_PROFILING_MEM', False):
                mem_status = 'memory: %4.2fM -> %4.2fM' % (start_mem, end_mem)
            else:
                mem_status = ''
            profile_logger.info('%s%s: %3.3fs. %s' % (level_prefix, func.__qualname__, elapsed, mem_status))
        else:
            r = func(*args, **kwargs)
        return r

    return arg_wrapper


class ColoredFormatter(Formatter):
    MAPPING = {
        'DEBUG': dict(color='white', on_color=None),  # white
        'INFO': dict(color='white', on_color=None),  # cyan
        'WARNING': dict(color='red', on_color='on_yellow'),  # yellow
        'ERROR': dict(color='white', on_color='on_red'),  # 31 for red
        'CRITICAL': dict(color='red', on_color='on_white'),  # white on red bg
        'SUCCESS': dict(color='white', on_color='on_green'),  # green
    }

    PREFIX = '\033['
    SUFFIX = '\033[0m'

    def __init__(self, pattern, **kwargs):
        super().__init__(pattern, **kwargs)

    def format(self, record):
        cr = copy(record)
        seq = self.MAPPING.get(cr.levelname, self.MAPPING['INFO'])  # default white
        cr.msg = colored(cr.msg, **seq)
        return super().format(cr)


def set_logger(context, verbose=False):
    if os.name == 'nt':  # for Windows
        return NTLogger(context, verbose)

    # Remove all handlers associated with the root logger object.
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    logger = logging.getLogger(context)
    logger.propagate = False
    if not logger.handlers:
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)
        formatter = ColoredFormatter(
            '%(levelname)-.1s:' + context + ':[%(filename).3s:%(funcName).3s:%(lineno)3d]:%(message)s', datefmt=
            '%m-%d %H:%M:%S')
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG if verbose else logging.INFO)
        console_handler.setFormatter(formatter)
        logger.handlers = []
        logger.addHandler(console_handler)

    return logger


class NTLogger:
    def __init__(self, context, verbose):
        self.context = context
        self.verbose = verbose

    def info(self, msg, **kwargs):
        print('I:%s:%s' % (self.context, msg), flush=True)

    def debug(self, msg, **kwargs):
        if self.verbose:
            print('D:%s:%s' % (self.context, msg), flush=True)

    def error(self, msg, **kwargs):
        print('E:%s:%s' % (self.context, msg), flush=True)

    def warning(self, msg, **kwargs):
        print('W:%s:%s' % (self.context, msg), flush=True)


class TimeContext:
    def __init__(self, msg):
        self._msg = msg

    def __enter__(self):
        self.start = time.perf_counter()
        print(self._msg, end=' ...\t', flush=True)

    def __exit__(self, typ, value, traceback):
        self.duration = time.perf_counter() - self.start
        print(colored('    [%3.3f secs]' % self.duration, 'green'), flush=True)


class Tokenizer:
    def __init__(self, dict_path: str = None):
        import jieba
        self._jieba = jieba.Tokenizer()
        self._jieba.cache_file = "gnes.jieba_wrapper.cache"

        if dict_path is not None:
            self._jieba.load_userdict(dict_path)

    def tokenize(self, text, with_position=False):
        if not with_position:
            return self._jieba.lcut(text)  # resulted token list
        else:
            return self._jieba.tokenize(text)  # triple data consisting of (token, start_pos, end_pos)


def batch_iterator(data: Union[Iterator[Any], List[Any], np.ndarray], batch_size: int, axis: int = 0) -> Iterator[Any]:
    if not batch_size or batch_size <= 0:
        yield data
        return
    if isinstance(data, np.ndarray):
        if batch_size >= data.shape[axis]:
            yield data
            return
        for _ in range(0, data.shape[axis], batch_size):
            start = _
            end = min(len(data), _ + batch_size)
            yield np.take(data, range(start, end), axis, mode='clip')
    elif hasattr(data, '__len__'):
        if batch_size >= len(data):
            yield data
            return
        for _ in range(0, len(data), batch_size):
            yield data[_:_ + batch_size]
    elif isinstance(data, Iterator):
        # as iterator, there is no way to know the length of it
        while True:
            chunk = tuple(islice(data, batch_size))
            if not chunk:
                return
            yield chunk
    else:
        raise TypeError('unsupported type: %s' % type(data))


def get_size(data: Union[Iterator[Any], List[Any], np.ndarray], axis: int = 0) -> int:
    if isinstance(data, np.ndarray):
        total_size = data.shape[axis]
    elif hasattr(data, '__len__'):
        total_size = len(data)
    else:
        total_size = None
    return total_size


def pooling_simple(data_array, pooling_strategy):
    if pooling_strategy == 'REDUCE_MEAN':
        _pooled_data = sum(data_array) / (len(data_array) + 1e-10)
    elif pooling_strategy == 'REDUCE_MAX':
        _pooled_data = max(data_array) / (len(data_array) + 1e-10)
    elif pooling_strategy == 'REDUCE_MEAN_MAX':
        _pooled_data = np.concatenate(
            (sum(data_array) / (len(data_array) + 1e-10),
             max(data_array) / (len(data_array) + 1e-10)), axis=0)
    else:
        raise ValueError('pooling_strategy: %s has not been implemented' % pooling_strategy)
    return _pooled_data


def pooling_np(data_array, pooling_strategy):
    if pooling_strategy == 'REDUCE_MEAN':
        _pooled_data = np.mean(data_array, axis=0)
    elif pooling_strategy == 'REDUCE_MAX':
        _pooled_data = np.amax(data_array, axis=0)
    elif pooling_strategy == 'REDUCE_MEAN_MAX':
        _pooled_data = np.concatenate(
            (np.mean(data_array, axis=0),
             np.amax(data_array, axis=0)), axis=1)
    else:
        raise ValueError('pooling_strategy: %s has not been implemented' % pooling_strategy)
    return _pooled_data


def pooling_torch(data_tensor, mask_tensor, pooling_strategy):
    import torch

    minus_mask = lambda x, m: x - (1.0 - m).unsqueeze(2) * 1e30
    mul_mask = lambda x, m: torch.mul(x, m.unsqueeze(2))

    masked_reduce_mean = lambda x, m: torch.div(torch.sum(mul_mask(x, m), dim=1),
                                                torch.sum(m.unsqueeze(2), dim=1) + 1e-10)
    masked_reduce_max = lambda x, m: torch.max(minus_mask(x, m), 1)[0]

    if pooling_strategy == 'REDUCE_MEAN':
        output_tensor = masked_reduce_mean(data_tensor, mask_tensor)
    elif pooling_strategy == 'REDUCE_MAX':
        output_tensor = masked_reduce_max(data_tensor, mask_tensor)
    elif pooling_strategy == 'REDUCE_MEAN_MAX':
        output_tensor = torch.cat(
            (masked_reduce_mean(data_tensor, mask_tensor),
             masked_reduce_max(data_tensor, mask_tensor)), dim=1)
    else:
        raise ValueError('pooling_strategy: %s has not been implemented' % pooling_strategy)

    return output_tensor


def batching(func: Callable[[Any], np.ndarray] = None, *,
             batch_size: Union[int, Callable] = None, num_batch=None,
             iter_axis: int = 0, concat_axis: int = 0):
    def _batching(func):
        @wraps(func)
        def arg_wrapper(self, data, label=None, *args, **kwargs):
            # priority: decorator > class_attribute
            b_size = (batch_size(data) if callable(batch_size) else batch_size) or getattr(self, 'batch_size', None)
            # no batching if b_size is None
            if b_size is None:
                if label is None:
                    return func(self, data, *args, **kwargs)
                else:
                    return func(self, data, label, *args, **kwargs)

            if hasattr(self, 'logger'):
                self.logger.info(
                    'batching enabled for %s(). batch_size=%s\tnum_batch=%s\taxis=%s' % (
                        func.__qualname__, b_size, num_batch, iter_axis))

            total_size1 = get_size(data, iter_axis)
            total_size2 = b_size * num_batch if num_batch else None

            if total_size1 is not None and total_size2 is not None:
                total_size = min(total_size1, total_size2)
            else:
                total_size = total_size1 or total_size2

            final_result = []

            if label is not None:
                data = (data, label)

            for b in batch_iterator(data[:total_size], b_size, iter_axis):
                if label is None:
                    r = func(self, b, *args, **kwargs)
                else:
                    r = func(self, b[0], b[1], *args, **kwargs)

                if r is not None:
                    final_result.append(r)

            if len(final_result) and concat_axis is not None and isinstance(final_result[0], np.ndarray):
                final_result = np.concatenate(final_result, concat_axis)

            if len(final_result):
                return final_result

        return arg_wrapper

    if func:
        return _batching(func)
    else:
        return _batching


def _get_yaml():
    y = YAML(typ='safe')
    y.default_flow_style = False
    return y


def parse_arg(v: str):
    if v.startswith('[') and v.endswith(']'):
        # function args must be immutable tuples not list
        tmp = v.replace('[', '').replace(']', '').strip().split(',')
        if len(tmp) > 0:
            return [parse_arg(vv.strip()) for vv in tmp]
        else:
            return []
    try:
        v = int(v)  # parse int parameter
    except ValueError:
        try:
            v = float(v)  # parse float parameter
        except ValueError:
            if len(v) == 0:
                # ignore it when the parameter is empty
                v = None
            elif v.lower() == 'true':  # parse boolean parameter
                v = True
            elif v.lower() == 'false':
                v = False
    return v


def countdown(t: int, logger=None, reason: str = 'I am blocking this thread'):
    if not logger:
        sys.stdout.write('\n')
        sys.stdout.flush()
    while t > 0:
        t -= 1
        msg = '%ss left: %s' % (colored('%3d' % t, 'yellow'), reason)
        if logger:
            logger.info(msg)
        else:
            sys.stdout.write('\r%s' % msg)
            sys.stdout.flush()
        time.sleep(1)
    sys.stdout.write('\n')
    sys.stdout.flush()


def train_required(func):
    @wraps(func)
    def arg_wrapper(self, *args, **kwargs):
        if hasattr(self, 'is_trained'):
            if self.is_trained:
                return func(self, *args, **kwargs)
            else:
                raise RuntimeError('training is required before calling "%s"' % func.__name__)
        else:
            raise AttributeError('%r has no attribute "is_trained"' % self)

    return arg_wrapper


def load_contrib_module():
    if not os.getenv('GNES_CONTRIB_MODULE_IS_LOADING'):
        import importlib.util

        contrib = os.getenv('GNES_CONTRIB_MODULE')
        os.environ['GNES_CONTRIB_MODULE_IS_LOADING'] = 'true'

        modules = []

        if contrib:
            default_logger.info(
                'find value in $GNES_CONTRIB_MODULE=%s, will try to load these modules from external' % contrib)
            for c in contrib.split(','):
                if ':' in c:
                    _name, _path = c.split(':')
                    spec = importlib.util.spec_from_file_location('gnes.contrib', _path)
                    foo = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(foo)
                    m = getattr(foo, _name)
                    modules.append(m)
                    default_logger.info('successfully register %s class, you can now use it via yaml.' % m)
        return modules


profile_logger = set_logger('PROFILE')
default_logger = set_logger('GNES')
profiling = time_profile

yaml = _get_yaml()



from os import path

from setuptools import setup, find_packages
from setuptools.extension import Extension

try:
    pkg_name = 'gnes'
    libinfo_py = path.join(pkg_name, '__init__.py')
    libinfo_content = open(libinfo_py, 'r').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][0]
    exec(version_line)  # produce __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

extensions = [
    Extension(
        'gnes.indexer.vector.bindexer.cython',
        ['gnes/indexer/vector/bindexer/bindexer.pyx'],
        extra_compile_args=['-O3'],
    ),
    Extension(
        'gnes.indexer.vector.hbindexer.cython',
        ['gnes/indexer/vector/hbindexer/hbindexer.pyx'],
        extra_compile_args=['-O3'],
    ),
]

base_dep = [
    'numpy',
    'scipy',
    'termcolor',
    'protobuf',
    'grpcio',
    'ruamel.yaml>=0.15.89',
    'pyzmq>=17.1.0',
]

extras_dep = {
    'bert': ['bert-serving-server>=1.8.6', 'bert-serving-client>=1.8.6'],
    # 'elmo': [
    #     'elmoformanylangs @ git+https://github.com/HIT-SCIR/ELMoForManyLangs.git@master#egg=elmoformanylangs-0.0.2',
    #     'paramiko', 'pattern3'],
    'flair': ['flair>=0.4.1'],
    'annoy': ['annoy==1.15.2'],
    'chinese': ['jieba'],
    'vision': ['opencv-python>=4.0.0', 'imagehash>=4.0'],
    'leveldb': ['plyvel>=1.0.5'],
    'test': ['pylint', 'memory_profiler>=0.55.0', 'psutil>=5.6.1', 'gputil>=1.4.0'],
    'http': ['flask', 'flask-compress', 'flask-cors', 'flask-json', 'aiohttp==3.5.4'],
    'onnx': ['onnxruntime'],
    'librosa': ['librosa>=0.7.0']
}


def combine_dep(new_key, base_keys):
    extras_dep[new_key] = list(set(k for v in base_keys for k in extras_dep[v]))


combine_dep('nlp', ['bert', 'flair'])
combine_dep('cn_nlp', ['chinese', 'nlp'])
combine_dep('all', [k for k in extras_dep if k != 'elmo'])

setup(
    name=pkg_name,
    packages=find_packages(),
    version=__version__,
    include_package_data=True,
    description='GNES is Generic Neural Elastic Search,'
                ' a cloud-native semantic search system based on deep neural network.',
    author='GNES team',
    author_email='team@gnes.ai',
    license='Apache 2.0',
    url='https://gnes.ai',
    download_url='https://github.com/gnes-ai/gnes/tags',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    setup_requires=[
        'setuptools>=18.0',
        'cython',
    ],
    ext_modules=extensions,
    install_requires=base_dep,
    extras_require=extras_dep,
    entry_points={
        'console_scripts': ['gnes=gnes.cli:main'],
    },
    classifiers=(
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Cython',
        'Programming Language :: Unix Shell',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Multimedia :: Video',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
    keywords='gnes cloud-native semantic search elastic neural-network encoding embedding serving',
)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import datetime
import io
import os
import subprocess as sp
from datetime import timedelta
from typing import List, Callable

import cv2
import numpy as np
from PIL import Image

from ..helper import set_logger

logger = set_logger(__name__, True)


def get_video_length(video_path):
    import re
    process = sp.Popen(['ffmpeg', '-i', video_path],
                       stdout=sp.PIPE,
                       stderr=sp.STDOUT)
    stdout, _ = process.communicate()
    stdout = str(stdout)
    matches = re.search(r"Duration:\s{1}(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),", stdout,
                        re.DOTALL).groupdict()
    h = float(matches['hours'])
    m = float(matches['minutes'])
    s = float(matches['seconds'])

    return 3600 * h + 60 * m + s


def get_video_length_from_raw(buffer_data):
    import re
    ffmpeg_cmd = ['ffmpeg', '-i', '-', '-']
    with sp.Popen(ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE,
                  bufsize=-1, shell=False) as pipe:
        _, stdout = pipe.communicate(buffer_data)
        stdout = stdout.decode()
        matches = re.search(r"Duration:\s{1}(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),", stdout,
                            re.DOTALL).groupdict()
        h = str(int(matches['hours']))
        m = str(int(matches['minutes']))
        s = str(round(float(matches['seconds'])))
        duration = datetime.datetime.strptime(h + ':' + m + ':' + s, '%H:%M:%S')
    return duration


def get_audio(buffer_data, sample_rate, interval,
              duration) -> List['np.ndarray']:
    import soundfile as sf

    audio_list = []
    start_time = datetime.datetime.strptime('00:00:00', '%H:%M:%S')
    while True:
        if start_time == duration:
            break

        end_time = start_time + timedelta(seconds=interval)
        if end_time > duration:
            end_time = duration
        ffmpeg_cmd = ['ffmpeg', '-i', '-',
                      '-f', 'wav',
                      '-ar', str(sample_rate),
                      '-ss', str(start_time).split(' ')[1],
                      '-to', str(end_time).split(' ')[1],
                      '-']

        # (-f, wav) output bytes in wav format
        # (-ar) sample rate
        # (-) output to stdout pipeline

        with sp.Popen(
                ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, bufsize=-1,
                shell=False) as pipe:
            raw_audio, _ = pipe.communicate(buffer_data)
            tmp_stream = io.BytesIO(raw_audio)
            data, sample_rate = sf.read(tmp_stream)
            # has multiple channels, do average
            if len(data.shape) == 2:
                data = np.mean(data, axis=1)
            audio_list.append(data)

        start_time = end_time

    return audio_list


def split_mp4_random(video_path, avg_length, max_clip_second=10):
    import random
    l = get_video_length(video_path)
    s = []
    num_part = max(int(l / avg_length), 2)

    while sum(s) < l:
        s.append(random.randint(3, max_clip_second))
    s[-1] = int(l - sum(s[:-1]))
    start = [sum(s[:i]) for i in range(len(s))]

    ts_group = [[] for _ in range(num_part)]

    for i, (_start, _du) in enumerate(zip(start, s)):
        ts_group[i % num_part].append(' -ss {} -t {} -i {} '.format(_start, _du, video_path))

    prefix = os.path.basename(video_path).replace('.mp4', '')
    for i in range(num_part):
        i_len = len(ts_group[i])
        cmd = 'ffmpeg' + ''.join(
            ts_group[i]) + '-filter_complex "{}concat=n={}:v=1:a=1" -strict -2 {}_{}.mp4 -y'.format(
            ''.join(['[{}]'.format(k) for k in range(i_len)]), i_len, prefix, i)
        os.system(cmd)


def get_video_frames(buffer_data: bytes, image_format: str = 'cv2',
                     **kwargs) -> List['np.ndarray']:
    ffmpeg_cmd = ['ffmpeg', '-i', '-', '-f', 'image2pipe']

    # example k,v pair:
    #    (-s, 420*360)
    #    (-vsync, vfr)
    #    (-vf, select=eq(pict_type\,I))
    for k, v in kwargs.items():
        if isinstance(v, (float, int)):
            v = str(v)
        ffmpeg_cmd.append('-' + k)
        ffmpeg_cmd.append(v)

    # (-c:v, png) output bytes in png format
    # (-an, -sn) disable audio processing
    # (-) output to stdout pipeline
    ffmpeg_cmd += ['-c:v', 'png', '-an', '-sn', '-']

    with sp.Popen(
            ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, bufsize=-1,
            shell=False) as pipe:
        stream, _ = pipe.communicate(buffer_data)

        # raw bytes for multiple PNGs.
        # split by PNG EOF b'\x89PNG'
        stream = stream.split(b'\x89PNG')

        if len(stream) <= 1:
            return []

        # reformulate the full pngs for feature processings.
        if image_format == 'pil':
            frames = [
                Image.open(io.BytesIO(b'\x89PNG' + _)) for _ in stream[1:]
            ]
        elif image_format == 'cv2':
            # frames = [
            #     cv2.imdecode(np.frombuffer(b'\x89PNG' + _, np.uint8), cv2.IMREAD_COLOR)
            #     for _ in stream[1:]
            # ]

            # TODO: to figure out the exception reason
            frames = []
            for _ in stream[1:]:
                image = cv2.imdecode(np.frombuffer(b'\x89PNG' + _, np.uint8), cv2.IMREAD_COLOR)
                try:
                    cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                    frames.append(image)
                except Exception as e:
                    logger.warning(
                        "The decoded cv2 image from keyframe buffer can not be converted to RGB: %s" % str(e))
        else:
            logger.error("The image format [%s] is not supported so far!" % image_format)
            raise NotImplementedError

    return frames


def block_descriptor(image: 'np.ndarray',
                     descriptor_fn: Callable,
                     num_blocks: int = 3) -> 'np.ndarray':
    h, w, _ = image.shape  # find shape of image and channel
    block_h = int(np.ceil(h / num_blocks))
    block_w = int(np.ceil(w / num_blocks))

    descriptors = []
    for i in range(0, h, block_h):
        for j in range(0, w, block_w):
            block = image[i:i + block_h, j:j + block_w]
            descriptors.extend(descriptor_fn(block))

    return np.array(descriptors)


def pyramid_descriptor(image: 'np.ndarray',
                       descriptor_fn: Callable,
                       max_level: int = 2) -> 'np.ndarray':
    descriptors = []
    for level in range(max_level + 1):
        num_blocks = 2 ** level
        descriptors.extend(block_descriptor(image, descriptor_fn, num_blocks))

    return np.array(descriptors)


def rgb_histogram(image: 'np.ndarray') -> 'np.ndarray':
    _, _, c = image.shape
    hist = [
        cv2.calcHist([image], [i], None, [256], [0, 256]) for i in range(c)
    ]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def hsv_histogram(image: 'np.ndarray') -> 'np.ndarray':
    _, _, c = image.shape
    hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)

    # sizes = [180, 256, 256]
    # ranges = [(0, 180), (0, 256), (0, 256)]

    # hist = [
    #     cv2.calcHist([hsv], [i], None, [sizes[i]], ranges[i]) for i in range(c)
    # ]

    hist = [cv2.calcHist([hsv], [i], None, [256], [0, 256]) for i in range(c)]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def phash_descriptor(image: 'np.ndarray'):
    image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
    import imagehash
    return imagehash.phash(image)


def compute_descriptor(image: 'np.ndarray',
                       method: str = 'rgb_histogram',
                       **kwargs) -> 'np.array':
    funcs = {
        'rgb_histogram': rgb_histogram,
        'hsv_histogram': hsv_histogram,
        'block_rgb_histogram': lambda image: block_descriptor(image, rgb_histogram, kwargs.get('num_blocks', 3)),
        'block_hsv_histogram': lambda image: block_descriptor(image, hsv_histogram, kwargs.get('num_blocks', 3)),
        'pyramid_rgb_histogram': lambda image: pyramid_descriptor(image, rgb_histogram, kwargs.get('max_level', 2)),
        'pyramid_hsv_histogram': lambda image: pyramid_descriptor(image, hsv_histogram, kwargs.get('max_level', 2)),
    }
    return funcs[method](image)


def compare_descriptor(descriptor1: 'np.ndarray',
                       descriptor2: 'np.ndarray',
                       metric: str = 'chisqr') -> float:
    dist_metric = {
        'correlation': cv2.HISTCMP_CORREL,
        'chisqr': cv2.HISTCMP_CHISQR,
        'chisqr_alt': cv2.HISTCMP_CHISQR_ALT,
        'intersection': cv2.HISTCMP_INTERSECT,
        'bhattacharya': cv2.HISTCMP_BHATTACHARYYA,
        'hellinguer': cv2.HISTCMP_HELLINGER,
        'kl_div': cv2.HISTCMP_KL_DIV
    }

    return cv2.compareHist(descriptor1, descriptor2, dist_metric[metric])



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import os
from typing import List

import numpy as np
from PIL import Image

from .base import BaseImagePreprocessor
from ...proto import array2blob


class SegmentPreprocessor(BaseImagePreprocessor):

    def __init__(self, model_name: str,
                 model_dir: str,
                 _use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name
        self.model_dir = model_dir
        self._use_cuda = _use_cuda

    def post_init(self):
        import torch
        import torchvision.models as models

        os.environ['TORCH_HOME'] = self.model_dir
        self._model = getattr(models.detection, self.model_name)(pretrained=True)
        self._model = self._model.eval()
        if self._use_cuda:
            # self._model.cuda()
            self._device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            self._model = self._model.to(self._device)

    def apply(self, doc: 'gnes_pb2.Document'):
        super().apply(doc)
        if doc.raw_bytes:
            original_image = Image.open(io.BytesIO(doc.raw_bytes))
            all_subareas, index = self._get_all_subarea(original_image)
            image_tensor = self._torch_transform(original_image)
            if self._use_cuda:
                image_tensor = image_tensor.cuda()

            seg_output = self._model([image_tensor])

            weight = seg_output[0]['scores'].tolist()
            length = len(list(filter(lambda x: x >= 0.5, weight)))
            chunks = seg_output[0]['boxes'].tolist()[:length]
            weight = weight[:length]

            for ci, ele in enumerate(zip(chunks, weight)):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.blob.CopyFrom(array2blob(self._crop_resize(original_image, ele[0])))
                c.offset_1d = ci
                c.offset_nd.x.extend(self._get_seg_offset_nd(all_subareas, index, ele[0]))
                c.weight = self._cal_area(ele[0]) / (original_image.size[0] * original_image.size[1])

            c = doc.chunks.add()
            c.doc_id = doc.doc_id
            c.blob.CopyFrom(array2blob(np.array(original_image.resize((self.target_img_size,
                                                                       self.target_img_size)))))
            c.offset_1d = len(chunks)
            c.offset_nd.x.extend([100, 100])
            c.weight = 1.
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def _crop_resize(self, original_image, coordinates):
        return np.array(original_image.crop(coordinates).resize((self.target_img_size,
                                                                 self.target_img_size)))

    def _get_seg_offset_nd(self, all_subareas: List[List[int]], index: List[List[int]], chunk: List[int]) -> List[int]:
        iou_list = [self._cal_iou(area, chunk) for area in all_subareas]
        return index[int(np.argmax(iou_list))][:2]

    @staticmethod
    def _cal_area(coordinate: List[int]):
        return (coordinate[2] - coordinate[0]) * (coordinate[3] - coordinate[1])

    def _cal_iou(self, image: List[int], chunk: List[int]) -> float:
        chunk_area = self._cal_area(chunk)
        image_area = self._cal_area(image)

        x1 = max(chunk[0], image[0])
        y1 = max(chunk[1], image[1])
        x2 = min(chunk[2], image[2])
        y2 = min(chunk[3], image[3])

        overlap_area = max(0, x2 - x1) * max(0, y2 - y1)
        iou = overlap_area / (chunk_area + image_area - overlap_area)
        return iou



import os
import unittest
from shutil import rmtree

from gnes.indexer.fulltext.leveldb import LVDBIndexer
from tests import txt_file2pb_docs


class TestBaseLVDB(unittest.TestCase):
    def setUp(self):
        dirname = os.path.dirname(__file__)

        self.test_docs = txt_file2pb_docs(open(os.path.join(dirname, 'tangshi.txt')))

        self.db_path = './test_leveldb'
        self.dump_path = os.path.join(dirname, 'indexer.bin')

    def tearDown(self):
        if os.path.exists(self.db_path):
            rmtree(self.db_path)

    def test_add_docs(self):
        db = LVDBIndexer(self.db_path)
        db.add(range(len(self.test_docs)), self.test_docs)
        self.assertTrue(os.path.exists(self.db_path))
        self.assertLess(0, len(os.listdir(self.db_path)))
        db.close()

    def test_query(self):
        db = LVDBIndexer(self.db_path)
        db.add(range(len(self.test_docs)), self.test_docs)

        res1 = db.query(range(len(self.test_docs)))
        num_non_empty = sum(1 for d in res1 if d)
        self.assertEqual(num_non_empty, len(self.test_docs))

        res2 = db.query(range(len(self.test_docs) + 1, len(self.test_docs) + 100))
        num_non_empty = sum(1 for d in res2 if d)
        self.assertEqual(num_non_empty, 0)
        db.close()

    def dump_load(self):
        tmp = LVDBIndexer(self.db_path)
        tmp.add(range(len(self.test_docs)), self.test_docs)
        tmp.dump(self.dump_path)
        tmp.close()

        db = LVDBIndexer.load(self.db_path)
        res1 = db.query(range(len(self.test_docs)))
        num_non_empty = sum(1 for d in res1 if d)
        self.assertEqual(num_non_empty, self.test_data1.length)

        res2 = db.query(range(len(self.test_docs) + 1, len(self.test_docs) + 100))
        num_non_empty = sum(1 for d in res2 if d)
        self.assertEqual(num_non_empty, 0)
        db.close()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import numpy as np

from ..base import BaseBinaryEncoder
from ...helper import batching, train_required


class PQEncoder(BaseBinaryEncoder):
    batch_size = 2048

    def __init__(self, num_bytes: int, cluster_per_byte: int = 255, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert 1 < cluster_per_byte <= 255, 'cluster number should >1 and <= 255 (0 is reserved for NOP)'
        self.num_bytes = num_bytes
        self.num_clusters = cluster_per_byte
        self.centroids = None

    def train(self, vecs: np.ndarray, *args, **kwargs):
        import faiss

        dim_per_byte = self._get_dim_per_byte(vecs)

        # use faiss ProductQuantizer directly
        vecs = vecs.astype(np.float32)
        model = faiss.ProductQuantizer(vecs.shape[1], self.num_bytes, 8)
        model.ksub = self.num_clusters
        model.byte_per_idx = 1

        model.train(vecs)
        centroids = faiss.vector_to_array(model.centroids)
        centroids = centroids[: self.num_clusters * vecs.shape[1]]
        self.centroids = np.reshape(centroids, [1, self.num_bytes,
                                                self.num_clusters,
                                                dim_per_byte])

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        dim_per_byte = self._get_dim_per_byte(vecs)

        x = np.reshape(vecs, [vecs.shape[0], self.num_bytes, 1, dim_per_byte])
        x = np.sum(np.square(x - self.centroids), -1)
        # start from 1
        x = np.argmax(-x, 2) + 1

        return np.array(x, dtype=np.uint8)

    def _get_dim_per_byte(self, vecs: np.ndarray):
        num_dim = vecs.shape[1]
        assert num_dim >= self.num_bytes and num_dim % self.num_bytes == 0, \
            'input dimension (=%d) should >= num_bytes (=%d) and can be divided by num_bytes!' % (
                num_dim, self.num_bytes)
        return int(num_dim / self.num_bytes)

    def _copy_from(self, x: 'PQEncoder') -> None:
        self.num_bytes = x.num_bytes
        self.num_clusters = x.num_clusters
        self.centroids = x.centroids
        self.is_trained = x.is_trained



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from typing import List
from typing import Optional

import numpy as np
import zmq

from . import gnes_pb2
from ..helper import batch_iterator

__all__ = ['send_message', 'recv_message', 'blob2array', 'array2blob', 'gnes_pb2', 'add_route']


class RequestGenerator:
    @staticmethod
    def index(data: List[bytes], batch_size: int = 0, start_id: int = 0, doc_type: int = gnes_pb2.Document.TEXT, *args, **kwargs):

        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = str(start_id)
            for raw_bytes in pi:
                d = req.index.docs.add()
                d.raw_bytes = raw_bytes
                d.weight = 1.0
                d.doc_type = doc_type
            yield req
            start_id += 1

    @staticmethod
    def train(data: List[bytes], batch_size: int = 0, start_id: int = 0, doc_type: int = gnes_pb2.Document.TEXT, *args, **kwargs):
        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = str(start_id)
            for raw_bytes in pi:
                d = req.train.docs.add()
                d.raw_bytes = raw_bytes
                d.doc_type = doc_type
            yield req
            start_id += 1
        req = gnes_pb2.Request()
        req.request_id = str(start_id)
        req.train.flush = True
        yield req
        start_id += 1

    @staticmethod
    def query(query: bytes, top_k: int, start_id: int = 0, doc_type: int = gnes_pb2.Document.TEXT, *args, **kwargs):
        if top_k <= 0:
            raise ValueError('"top_k: %d" is not a valid number' % top_k)

        req = gnes_pb2.Request()
        req.request_id = str(start_id)
        req.search.query.raw_bytes = query
        req.search.query.doc_type = doc_type
        req.search.top_k = top_k
        yield req


def blob2array(blob: 'gnes_pb2.NdArray') -> np.ndarray:
    """
    Convert a blob proto to an array.
    """
    return np.frombuffer(blob.data, dtype=blob.dtype).reshape(blob.shape)


def array2blob(x: np.ndarray) -> 'gnes_pb2.NdArray':
    """Converts a N-dimensional array to blob proto.
    """
    blob = gnes_pb2.NdArray()
    blob.data = x.tobytes()
    blob.shape.extend(list(x.shape))
    blob.dtype = x.dtype.name
    return blob


def new_envelope(client_id: str,
                 request_id: str = str(uuid.uuid4()),
                 num_part: int = 1,
                 part_id: int = 1,
                 timeout: int = 5000) -> 'gnes_pb2.Envelope':
    evlp = gnes_pb2.Envelope()
    evlp.client_id = client_id
    evlp.request_id = request_id
    evlp.part_id = part_id
    evlp.num_part.append(num_part)
    evlp.timeout = timeout
    add_route(evlp, client_id)
    return evlp


def add_route(evlp: 'gnes_pb2.Envelope', name: str):
    r = evlp.routes.add()
    r.service = name
    r.timestamp.GetCurrentTime()


def merge_routes(msg: 'gnes_pb2.Message', prev_msgs: List['gnes_pb2.Message'], idx: int = -1):
    msg.envelope.routes.extend([m.envelope.routes[idx] for m in prev_msgs[:-1]])


def send_message(sock: 'zmq.Socket', msg: 'gnes_pb2.Message', timeout: int = -1) -> None:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.SNDTIMEO, timeout)
        else:
            sock.setsockopt(zmq.SNDTIMEO, -1)

        sock.send_multipart([msg.envelope.client_id.encode(), msg.SerializeToString()])
    except zmq.error.Again:
        raise TimeoutError(
            'cannot send message to sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    finally:
        sock.setsockopt(zmq.SNDTIMEO, -1)


def recv_message(sock: 'zmq.Socket', timeout: int = -1) -> Optional['gnes_pb2.Message']:
    response = []
    try:
        if timeout > 0:
            sock.setsockopt(zmq.RCVTIMEO, timeout)
        else:
            sock.setsockopt(zmq.RCVTIMEO, -1)

        _, msg_data = sock.recv_multipart()
        msg = gnes_pb2.Message()
        msg.ParseFromString(msg_data)
        return msg

    except ValueError:
        raise ValueError('received a wrongly-formatted request (expected 4 frames, got %d)' % len(response))
    except zmq.error.Again:
        raise TimeoutError(
            'no response from sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    finally:
        sock.setsockopt(zmq.RCVTIMEO, -1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import zipfile

import grpc

from ..proto import gnes_pb2_grpc, RequestGenerator


class CLIClient:
    def __init__(self, args):
        if args.data_type == 'text':
            all_bytes = [v.encode() for v in args.txt_file]
        elif args.data_type == 'image':
            zipfile_ = zipfile.ZipFile(args.image_zip_file, "r")
            all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
        elif args.data_type == 'video':
            zipfile_ = zipfile.ZipFile(args.video_zip_file, "r")
            all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
        else:
            raise ValueError(
                '--data_type can only be text, image or video')

        with grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', args.max_message_size * 1024 * 1024),
                         ('grpc.max_receive_message_length', args.max_message_size * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)

            if args.mode == 'train':
                resp = list(stub.StreamCall(RequestGenerator.train(all_bytes, args.batch_size)))[-1]
                print(resp)
            elif args.mode == 'index':
                resp = list(stub.StreamCall(RequestGenerator.train(all_bytes, args.batch_size)))[-1]
                print(resp)
            elif args.mode == 'query':
                for idx, q in enumerate(all_bytes):
                    for req in RequestGenerator.query(q, args.top_k):
                        resp = stub.Call(req)
                        print(resp)
                        print('query %d result: %s' % (idx, resp))
                        input('press any key to continue...')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
from PIL import Image

from .base import BaseImagePreprocessor
from ...proto import gnes_pb2, blob2array, array2blob


class ResizeChunkPreprocessor(BaseImagePreprocessor):

    def __init__(self,
                 target_width: int = 224,
                 target_height: int = 224,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_width = target_width
        self.target_height = target_height

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        for c in doc.chunks:
            img = blob2array(c.blob)
            img = np.array(Image.fromarray(img).resize((self.target_width, self.target_height)))
            c.blob.CopyFrom(array2blob(img))



import os
import unittest
from concurrent.futures import ThreadPoolExecutor

import grpc

from gnes.cli.parser import set_grpc_service_parser, set_frontend_parser
from gnes.proto import gnes_pb2_grpc, RequestGenerator
from gnes.service.frontend import FrontendService
from gnes.service.grpc import GRPCService
from tests.proto_s import dummy_pb2_grpc


class DummyServer:
    def __init__(self, bind_address):
        self.server = grpc.server(
            ThreadPoolExecutor(max_workers=1),
            options=[('grpc.max_send_message_length', 1 * 1024 * 1024),
                     ('grpc.max_receive_message_length', 1 * 1024 * 1024)])
        dummy_pb2_grpc.add_DummyGRPCServiceServicer_to_server(self.GNESServicer(), self.server)
        self.bind_address = bind_address
        self.server.add_insecure_port(self.bind_address)

    def __enter__(self):
        self.server.start()
        print('dummy server is listening at: %s' % self.bind_address)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.stop(None)

    class GNESServicer(dummy_pb2_grpc.DummyGRPCServiceServicer):

        def dummyAPI(self, request, context):
            print('the dummy server received something: %s' % request)
            return request


class TestGRPCService(unittest.TestCase):
    def setUp(self):
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')
        self.dirname = os.path.dirname(__file__)

        self.s_args = set_grpc_service_parser().parse_args([
            '--grpc_port', '5678',
            '--grpc_host', '127.0.0.1',
            '--pb2_path', os.path.join(self.dirname, 'proto', 'dummy_pb2.py'),
            '--pb2_grpc_path', os.path.join(self.dirname, 'proto', 'dummy_pb2_grpc.py'),
            '--stub_name', 'DummyGRPCServiceStub',
            '--api_name', 'dummyAPI'
        ])

        self.args = set_frontend_parser().parse_args([
            '--grpc_host', '127.0.0.1',
            '--grpc_port', '9999',
            '--port_in', str(self.s_args.port_out),
            '--port_out', str(self.s_args.port_in),
        ])

    def test_grpc_empty_service(self):
        with DummyServer('%s:%d' % (self.s_args.grpc_host, self.s_args.grpc_port)), GRPCService(self.s_args):
            pass

    @unittest.SkipTest
    def test_grpc_real_service(self):
        # to fix

        with DummyServer('%s:%d' % (self.s_args.grpc_host, self.s_args.grpc_port)), GRPCService(
                self.s_args), FrontendService(self.args), grpc.insecure_channel(
            '%s:%s' % (self.args.grpc_host, self.args.grpc_port),
            options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                     ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            resp = stub.Call(list(RequestGenerator.query(b'abc', 1))[0])
            self.assertEqual(resp.request_id, '0')  # idx start with 0, but +1 for final FLUSH



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import random
import time
from collections import defaultdict
from typing import Dict, List

from pkg_resources import resource_stream
from ruamel.yaml import YAML, StringIO
from ruamel.yaml.comments import CommentedMap

from .. import __version__
from ..cli.parser import set_frontend_parser, \
    set_router_service_parser, set_loadable_service_parser, set_preprocessor_service_parser, \
    set_indexer_service_parser
from ..helper import set_logger
from ..service.base import SocketType

_yaml = YAML()


class YamlComposer:
    comp2file = {
        'Encoder': 'encode',
        'Router': 'route',
        'Indexer': 'index',
        'Frontend': 'frontend',
        'Preprocessor': 'preprocess'
    }

    comp2args = {
        'Encoder': set_loadable_service_parser().parse_args(['--yaml_path', 'BaseEncoder']),
        'Router': set_router_service_parser().parse_args(['--yaml_path', 'BaseRouter']),
        'Indexer': set_indexer_service_parser().parse_args(['--yaml_path', 'BaseIndexer']),
        'Frontend': set_frontend_parser().parse_args([]),
        'Preprocessor': set_preprocessor_service_parser().parse_args(['--yaml_path', 'BasePreprocessor'])
    }

    class Layer:
        default_values = {
            'name': None,
            'yaml_path': None,
            'replicas': 1,
            'income': 'pull'
        }

        def __init__(self, layer_id: int = 0):
            self.layer_id = layer_id
            self.components = []

        @staticmethod
        def get_value(comp: Dict, key: str):
            return comp.get(key, YamlComposer.Layer.default_values[key])

        @property
        def is_homogenous(self):
            return len(self.components) == 1

        @property
        def is_single_component(self):
            return self.is_homogenous and self.get_value(self.components[0], 'replicas') == 1

        @property
        def is_homo_multi_component(self):
            return self.is_homogenous and not self.is_single_component

        @property
        def is_heto_single_component(self):
            return not self.is_homogenous and all(self.get_value(c, 'replicas') == 1 for c in self.components)

        @property
        def get_component_name(self, unique: bool = False):
            r = [c['name'] for c in self.components]
            if unique:
                r = list(set(r))
            return r

        def append(self, comp):
            self.components.append(comp)

        def __repr__(self):
            return str(self.components)

    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)

        with args.yaml_path:
            tmp = _yaml.load(args.yaml_path)
            stream = StringIO()
            _yaml.dump(tmp, stream)
            self.original_yaml = stream.getvalue().strip()

        self._name = tmp.get('name', args.name)
        self._port = tmp.get('port', args.port)
        self._networks = tmp.get('networks', {})
        self._volumes = tmp.get('volumes', {})

        self._layers = []  # type: List['YamlComposer.Layer']
        self._num_layer = 0

        if 'services' in tmp:
            self.add_layer()
            self.add_comp(CommentedMap({'name': 'Frontend', 'grpc_port': self._port}))
            for comp in tmp['services']:
                self.add_layer()
                if isinstance(comp, list):
                    for c in comp:
                        self.add_comp(c)
                elif self.check_fields(comp):
                    self.add_comp(comp)
                else:
                    raise ValueError(comp)
        else:
            self.logger.error('yaml file defines an empty graph! no "component" field exists!')

    def check_fields(self, comp: Dict) -> bool:
        if 'name' not in comp:
            raise AttributeError('a component must have a name (choices: %s)' % ', '.join(self.comp2file.keys()))
        if comp['name'] not in self.comp2file:
            raise AttributeError(
                'a component must be one of: %s, but given %s' % (', '.join(self.comp2file.keys()), comp['name']))
        if 'yaml_path' not in comp:
            self.logger.warning(
                'found empty "yaml_path", '
                'i will use a default config and would probably result in an empty model')
        for k in comp:
            if k not in self.Layer.default_values:
                self.logger.warning('your yaml contains an unrecognized key named "%s"' % k)
        return True

    def add_layer(self, layer: 'Layer' = None) -> None:
        self._layers.append(copy.deepcopy(layer) or self.Layer(layer_id=self._num_layer))
        self._num_layer += 1

    def add_comp(self, comp: Dict) -> None:
        self._layers[-1].append(comp)

    def build_layers(self) -> List['YamlComposer.Layer']:
        all_layers = []  # type: List['YamlComposer.Layer']
        for idx, layer in enumerate(self._layers[1:] + [self._layers[0]], 1):
            last_layer = self._layers[idx - 1]
            for l in self._add_router(last_layer, layer):
                all_layers.append(copy.deepcopy(l))
        all_layers[0] = copy.deepcopy(self._layers[0])

        # Frontend should always on the bind role
        assert all_layers[0].is_single_component
        assert all_layers[0].components[0]['name'] == 'Frontend'

        if all_layers[0].components[0]['socket_in'] == str(SocketType.SUB_CONNECT):
            # change to sub bind
            all_layers[0].components[0]['socket_in'] = str(SocketType.SUB_BIND)
            for c in all_layers[-1].components:
                c['socket_out'] = str(SocketType.PUB_CONNECT)

        if all_layers[0].components[0]['socket_in'] == str(SocketType.PULL_CONNECT):
            # change to sub bind
            all_layers[0].components[0]['socket_in'] = str(SocketType.PULL_BIND)
            for c in all_layers[-1].components:
                c['socket_out'] = str(SocketType.PUSH_CONNECT)

        return all_layers

    @staticmethod
    def build_dockerswarm(all_layers: List['YamlComposer.Layer'], docker_img: str = 'gnes/gnes:latest',
                          volumes: Dict = None, networks: Dict = None) -> str:
        with resource_stream('gnes', '/'.join(('resources', 'compose', 'gnes-swarm.yml'))) as r:
            swarm_lines = _yaml.load(r)
        config_dict = {}
        for l_idx, layer in enumerate(all_layers):
            for c_idx, c in enumerate(layer.components):
                c_name = '%s%d%d' % (c['name'], l_idx, c_idx)
                args = ['--%s %s' % (a, str(v) if ' ' not in str(v) else ('"%s"' % str(v))) for a, v in c.items() if
                        a in YamlComposer.comp2args[c['name']] and a != 'yaml_path' and v]
                if 'yaml_path' in c and c['yaml_path'] is not None:
                    if c['yaml_path'].endswith('.yml') or c['yaml_path'].endswith('.yaml'):
                        args.append('--yaml_path /%s_yaml' % c_name)
                        config_dict['%s_yaml' % c_name] = {'file': c['yaml_path']}
                    else:
                        args.append('--yaml_path %s' % c['yaml_path'])

                if l_idx + 1 < len(all_layers):
                    next_layer = all_layers[l_idx + 1]
                    _l_idx = l_idx + 1
                else:
                    next_layer = all_layers[0]
                    _l_idx = 0
                host_out_name = ''
                for _c_idx, _c in enumerate(next_layer.components):
                    if _c['port_in'] == c['port_out']:
                        host_out_name = '%s%d%d' % (_c['name'], _l_idx, _c_idx)
                        break

                if l_idx - 1 >= 0:
                    last_layer = all_layers[l_idx - 1]
                    _l_idx = l_idx - 1
                else:
                    last_layer = all_layers[-1]
                    _l_idx = len(all_layers) - 1

                host_in_name = ''
                for _c_idx, _c in enumerate(last_layer.components):
                    if _c['port_out'] == c['port_in']:
                        host_in_name = '%s%d%d' % (_c['name'], _l_idx, _c_idx)
                        break

                if 'BIND' not in c['socket_out']:
                    args.append('--host_out %s' % host_out_name)
                if 'BIND' not in c['socket_in']:
                    args.append('--host_in %s' % host_in_name)

                cmd = '%s %s' % (YamlComposer.comp2file[c['name']], ' '.join(args))
                swarm_lines['services'][c_name] = CommentedMap({
                    'image': docker_img,
                    'command': cmd,
                })

                rep_c = YamlComposer.Layer.get_value(c, 'replicas')
                if rep_c > 1:
                    swarm_lines['services'][c_name]['deploy'] = CommentedMap({
                        'replicas': YamlComposer.Layer.get_value(c, 'replicas'),
                        'restart_policy': {
                            'condition': 'on-failure',
                            'max_attempts': 3,
                        }
                    })

                if 'yaml_path' in c and c['yaml_path'] is not None \
                        and (c['yaml_path'].endswith('.yml') or c['yaml_path'].endswith('.yaml')):
                    swarm_lines['services'][c_name]['configs'] = ['%s_yaml' % c_name]

                if c['name'] == 'Frontend':
                    swarm_lines['services'][c_name]['ports'] = ['%d:%d' % (c['grpc_port'], c['grpc_port'])]

        if volumes:
            swarm_lines['volumes'] = volumes
        if networks:
            swarm_lines['networks'] = volumes
        swarm_lines['configs'] = config_dict
        stream = StringIO()
        _yaml.dump(swarm_lines, stream)
        return stream.getvalue().strip()

    @staticmethod
    def build_kubernetes(all_layers: List['YamlComposer.Layer'], *args, **kwargs):
        pass

    @staticmethod
    def build_shell(all_layers: List['YamlComposer.Layer'], log_redirect: str = None) -> str:
        shell_lines = []
        for layer in all_layers:
            for c in layer.components:
                rep_c = YamlComposer.Layer.get_value(c, 'replicas')
                shell_lines.append('printf "starting service \\e[1;33m%s\\e[0m with \e[1;33m%s\e[0m replicas...\\n"' % (
                    c['name'], rep_c))
                for _ in range(rep_c):
                    cmd = YamlComposer.comp2file[c['name']]
                    args = ' '.join(
                        ['--%s %s' % (a, str(v) if ' ' not in str(v) else ('"%s"' % str(v))) for a, v in c.items() if
                         a in YamlComposer.comp2args[c['name']] and v])
                    shell_lines.append('gnes %s %s %s &' % (
                        cmd, args, '>> %s 2>&1' % log_redirect if log_redirect else ''))

        with resource_stream('gnes', '/'.join(('resources', 'compose', 'gnes-shell.sh'))) as r:
            return r.read().decode().replace('{{gnes-template}}', '\n'.join(shell_lines)).strip()

    @staticmethod
    def build_mermaid(all_layers: List['YamlComposer.Layer'], mermaid_leftright: bool = False) -> str:
        mermaid_graph = []
        cls_dict = defaultdict(set)
        for l_idx, layer in enumerate(all_layers[1:] + [all_layers[0]], 1):
            last_layer = all_layers[l_idx - 1]

            for c_idx, c in enumerate(last_layer.components):
                # if len(last_layer.components) > 1:
                #     self.mermaid_graph.append('\tsubgraph %s%d' % (c['name'], c_idx))
                for j in range(YamlComposer.Layer.get_value(c, 'replicas')):
                    for c1_idx, c1 in enumerate(layer.components):
                        if c1['port_in'] == c['port_out']:
                            p = '((%s%s))' if c['name'] == 'Router' else '(%s%s)'
                            p1 = '((%s%s))' if c1['name'] == 'Router' else '(%s%s)'
                            for j1 in range(YamlComposer.Layer.get_value(c1, 'replicas')):
                                _id, _id1 = '%s%s%s' % (last_layer.layer_id, c_idx, j), '%s%s%s' % (
                                    layer.layer_id, c1_idx, j1)
                                conn_type = (
                                        c['socket_out'].split('_')[0] + '/' + c1['socket_in'].split('_')[0]).lower()
                                s_id = '%s%s' % (c_idx if len(last_layer.components) > 1 else '',
                                                 j if YamlComposer.Layer.get_value(c, 'replicas') > 1 else '')
                                s1_id = '%s%s' % (c1_idx if len(layer.components) > 1 else '',
                                                  j1 if YamlComposer.Layer.get_value(c1, 'replicas') > 1 else '')
                                mermaid_graph.append(
                                    '\t%s%s%s-- %s -->%s%s%s' % (
                                        c['name'], _id, p % (c['name'], s_id), conn_type, c1['name'], _id1,
                                        p1 % (c1['name'], s1_id)))
                                cls_dict[c['name'] + 'CLS'].add('%s%s' % (c['name'], _id))
                                cls_dict[c1['name'] + 'CLS'].add('%s%s' % (c1['name'], _id1))
                # if len(last_layer.components) > 1:
                #     self.mermaid_graph.append('\tend')

        style = ['classDef FrontendCLS fill:#FFE0E0,stroke:#FFE0E0,stroke-width:1px;',
                 'classDef EncoderCLS fill:#FFDAAF,stroke:#FFDAAF,stroke-width:1px;',
                 'classDef IndexerCLS fill:#FFFBC1,stroke:#FFFBC1,stroke-width:1px;',
                 'classDef RouterCLS fill:#C9E8D2,stroke:#C9E8D2,stroke-width:1px;',
                 'classDef PreprocessorCLS fill:#CEEEEF,stroke:#CEEEEF,stroke-width:1px;']
        class_def = ['class %s %s;' % (','.join(v), k) for k, v in cls_dict.items()]
        mermaid_str = '\n'.join(
            ['graph %s' % ('LR' if mermaid_leftright else 'TD')] + mermaid_graph + style + class_def)
        return mermaid_str.strip()

    @staticmethod
    def build_html(generate_dict: Dict[str, str]) -> str:
        with resource_stream('gnes', '/'.join(('resources', 'compose', 'gnes-board.html'))) as r:
            html = r.read().decode()
            for k, v in generate_dict.items():
                if v:
                    html = html.replace('{{gnes-%s}}' % k, v)
        return html.strip()

    def build_all(self):
        def std_or_print(f, content):
            if content:
                if f:
                    with f as fp:
                        fp.write(content)
                        self.logger.info('generated content is written to %s' % f)
                else:
                    self.logger.warning('no file path is defined, i will just print it to stdout')
                    print(content)

        all_layers = self.build_layers()
        cmds = {
            'mermaid': self.build_mermaid(all_layers, self.args.mermaid_leftright),
            'shell': self.build_shell(all_layers, self.args.shell_log_redirect),
            'yaml': self.original_yaml,
            'docker': self.build_dockerswarm(all_layers, self.args.docker_img,
                                             volumes=self._volumes, networks=self._networks),
            'k8s': self.build_kubernetes(all_layers),
            'timestamp': time.strftime("%a, %d %b %Y %H:%M:%S"),
            'version': __version__
        }

        cmds['html'] = self.build_html(cmds)

        std_or_print(self.args.graph_path, cmds['mermaid'])
        std_or_print(self.args.shell_path, cmds['shell'])
        std_or_print(self.args.swarm_path, cmds['docker'])
        std_or_print(self.args.k8s_path, cmds['k8s'])
        std_or_print(self.args.html_path, cmds['html'])
        return cmds

    @staticmethod
    def _get_random_port(min_port: int = 49152, max_port: int = 65536) -> str:
        return str(random.randrange(min_port, max_port))

    @staticmethod
    def _get_random_host(comp_name: str) -> str:
        return str(comp_name + str(random.randrange(0, 100)))

    def _add_router(self, last_layer: 'YamlComposer.Layer', layer: 'YamlComposer.Layer') -> List['YamlComposer.Layer']:
        def rule1():
            # a shortcut fn: push connect the last and current
            last_layer.components[0]['socket_out'] = str(SocketType.PUSH_BIND)
            layer.components[0]['socket_in'] = str(SocketType.PULL_CONNECT)

        def rule2():
            # a shortcut fn: pub connect the last and the current
            last_layer.components[0]['socket_out'] = str(SocketType.PUB_BIND)
            layer.components[0]['socket_in'] = str(SocketType.SUB_CONNECT)

        def rule3():
            # a shortcut fn: (N)-2-(N) with push pull connection
            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            last_layer.components[0]['socket_out'] = str(SocketType.PUSH_CONNECT)
            r = CommentedMap({'name': 'Router',
                              'yaml_path': 'BaseRouter',
                              'socket_in': str(SocketType.PULL_BIND),
                              'socket_out': str(SocketType.PUSH_BIND),
                              'port_in': last_layer.components[0]['port_out'],
                              'port_out': self._get_random_port()})
            for c in layer.components:
                c['socket_in'] = str(SocketType.PULL_CONNECT)
                c['port_in'] = r['port_out']
            router_layer.append(r)
            router_layers.append(router_layer)

        def rule4():
            # a shortcut fn: (N)-to-(1)&(1)&(1)
            last_layer.components[0]['socket_out'] = str(SocketType.PUB_BIND)
            for c in layer.components:
                c['socket_in'] = str(SocketType.SUB_CONNECT)

        def rule5():
            # a shortcut fn: based on c3(): (N)-2-(N) with pub sub connection
            rule3()
            router_layers[0].components[0]['socket_out'] = str(SocketType.PUB_BIND)
            router_layers[0].components[0]['yaml_path'] = '"!PublishRouter {parameter: {num_part: %d}}"' \
                                                          % len(layer.components)
            for c in layer.components:
                c['socket_in'] = str(SocketType.SUB_CONNECT)

        def rule6():
            last_layer.components[0]['socket_out'] = str(SocketType.PUB_BIND)
            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            for c in layer.components:
                income = self.Layer.get_value(c, 'income')
                r = CommentedMap({'name': 'Router',
                                  'yaml_path': 'BaseReduceRouter',
                                  'socket_in': str(SocketType.SUB_CONNECT),
                                  'socket_out': str(SocketType.PUSH_BIND) if income == 'pull' else str(
                                      SocketType.PUB_BIND),
                                  'port_in': last_layer.components[0]['port_out'],
                                  'port_out': self._get_random_port()})
                c['socket_in'] = str(SocketType.PULL_CONNECT) if income == 'pull' else str(SocketType.SUB_CONNECT)
                c['port_in'] = r['port_out']
                router_layer.append(r)
            router_layers.append(router_layer)

        def rule7():
            last_layer.components[0]['socket_out'] = str(SocketType.PUSH_CONNECT)

            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            r0 = CommentedMap({'name': 'Router',
                               'yaml_path': '"!PublishRouter {parameter: {num_part: %d}}"' % len(layer.components),
                               'socket_in': str(SocketType.PULL_BIND),
                               'socket_out': str(SocketType.PUB_BIND),
                               'port_in': self._get_random_port(),
                               'port_out': self._get_random_port()})
            router_layer.append(r0)
            router_layers.append(router_layer)
            last_layer.components[0]['port_out'] = r0['port_in']

            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            for c in layer.components:
                r = CommentedMap({'name': 'Router',
                                  'yaml_path': 'BaseRouter',
                                  'socket_in': str(SocketType.SUB_CONNECT),
                                  'socket_out': str(SocketType.PUSH_BIND),
                                  'port_in': r0['port_out'],
                                  'port_out': self._get_random_port()})
                c['socket_in'] = str(SocketType.PULL_CONNECT)
                c['port_in'] = r['port_out']
                router_layer.append(r)
            router_layers.append(router_layer)

        def rule10():
            last_layer.components[0]['socket_out'] = str(SocketType.PUSH_CONNECT)

            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            r0 = CommentedMap({'name': 'Router',
                               'yaml_path': '"!PublishRouter {parameter: {num_part: %d}}"' % len(layer.components),
                               'socket_in': str(SocketType.PULL_BIND),
                               'socket_out': str(SocketType.PUB_BIND),
                               'port_in': self._get_random_port(),
                               'port_out': self._get_random_port()})
            router_layer.append(r0)
            router_layers.append(router_layer)
            last_layer.components[0]['port_out'] = r0['port_in']

            for c in layer.components:
                c['socket_in'] = str(SocketType.SUB_CONNECT)
                c['port_in'] = r0['port_out']

        def rule8():
            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            r = CommentedMap({'name': 'Router',
                              'yaml_path': 'BaseReduceRouter',
                              'socket_in': str(SocketType.PULL_BIND),
                              'socket_out': str(SocketType.PUSH_BIND),
                              'port_in': self._get_random_port(),
                              'port_out': self._get_random_port()})

            for c in last_layer.components:
                last_income = self.Layer.get_value(c, 'income')
                if last_income == 'sub':
                    c['socket_out'] = str(SocketType.PUSH_CONNECT)
                    r_c = CommentedMap({'name': 'Router',
                                        'yaml_path': 'BaseReduceRouter',
                                        'socket_in': str(SocketType.PULL_BIND),
                                        'socket_out': str(SocketType.PUSH_CONNECT),
                                        'port_in': self._get_random_port(),
                                        'port_out': r['port_in']})
                    c['port_out'] = r_c['port_in']
                    router_layer.append(r_c)
                elif last_income == 'pull':
                    c['socket_out'] = str(SocketType.PUSH_CONNECT)
                    c['port_out'] = r['port_in']

            for c in layer.components:
                c['socket_in'] = str(SocketType.PULL_CONNECT)
                c['port_in'] = r['port_out']

            if router_layer.components:
                router_layers.append(router_layer)
            else:
                self._num_layer -= 1

            router_layer = YamlComposer.Layer(layer_id=self._num_layer)
            self._num_layer += 1
            router_layer.append(r)
            router_layers.append(router_layer)

        def rule9():
            # a shortcut fn: push connect the last and current
            last_layer.components[0]['socket_out'] = str(SocketType.PUSH_CONNECT)
            layer.components[0]['socket_in'] = str(SocketType.PULL_BIND)

        router_layers = []  # type: List['self.Layer']
        # bind the last out to current in

        if last_layer.is_single_component:
            last_layer.components[0]['port_out'] = self._get_random_port()
            for c in layer.components:
                c['port_in'] = last_layer.components[0]['port_out']
            # 1-to-?
            if layer.is_single_component:
                # 1-to-(1)
                # no router is needed
                rule1()
            elif layer.is_homo_multi_component:
                # 1-to-(N)
                income = self.Layer.get_value(layer.components[0], 'income')
                if income == 'pull':
                    rule1()
                elif income == 'sub':
                    rule2()
                else:
                    raise NotImplementedError('replica type: %s is not recognized!' % income)
            elif layer.is_heto_single_component:
                # 1-to-(1)&(1)&(1)
                rule4()
            else:
                # 1-to-(N)&(N)&(N)
                rule6()
        elif last_layer.is_homo_multi_component:
            # (N)-to-?
            last_layer.components[0]['port_out'] = self._get_random_port()

            last_income = self.Layer.get_value(last_layer.components[0], 'income')

            for c in layer.components:
                c['port_in'] = last_layer.components[0]['port_out']

            if layer.is_single_component:
                if last_income == 'pull':
                    # (N)-to-1
                    rule9()
                elif last_income == 'sub':
                    # (N)-to-1 with a sync barrier
                    rule3()
                else:
                    raise NotImplementedError('replica type: %s is not recognized!' % last_income)
            elif layer.is_homo_multi_component:
                # (N)-to-(N)
                # need a router anyway
                if self.Layer.get_value(layer.components[0], 'income') == 'sub':
                    rule5()
                else:
                    rule3()
            elif layer.is_heto_single_component:
                # (N)-to-(1)&(1)&(1)
                rule5()
            else:
                income = self.Layer.get_value(layer.components[0], 'income')
                if income == 'pull':
                    rule7()
                elif income == 'sub':
                    rule10()
                else:
                    raise NotImplementedError('replica type: %s is not recognized!' % last_income)
        elif last_layer.is_heto_single_component:
            rule8()
        else:
            rule8()
        return [last_layer, *router_layers]



import os
import unittest

import numpy as np

from gnes.cli.parser import set_router_service_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2, array2blob
from gnes.service.base import SocketType
from gnes.service.router import RouterService


class TestProto(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.publish_router_yaml = '!PublishRouter {parameter: {num_part: 2}}'
        self.batch_router_yaml = '!DocBatchRouter {gnes_config: {batch_size: 2}}'
        self.reduce_router_yaml = 'BaseReduceRouter'
        self.chunk_router_yaml = 'ChunkToDocumentRouter'
        self.chunk_sum_yaml = 'ChunkSumRouter'
        self.doc_router_yaml = 'DocFillRouter'
        self.doc_sum_yaml = 'DocSumRouter'
        self.concat_router_yaml = 'ConcatEmbedRouter'

    def test_service_empty(self):
        args = set_router_service_parser().parse_args(['--yaml_path', 'BaseRouter'])
        with RouterService(args):
            pass

    def test_map_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.batch_router_yaml,
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertEqual(len(r.request.index.docs), 2)
            r = c1.recv_message()
            self.assertEqual(len(r.request.index.docs), 2)
            r = c1.recv_message()
            self.assertEqual(len(r.request.index.docs), 1)

    def test_publish_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.publish_router_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1, ZmqClient(c_args) as c2:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            msg.envelope.num_part.append(1)
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1, 2])
            r = c2.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1, 2])

    def test_reduce_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.reduce_router_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1, ZmqClient(c_args) as c2:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            msg.envelope.num_part.extend([1, 3])
            c1.send_message(msg)
            c1.send_message(msg)
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            print(r.envelope.routes)

    def test_chunk_reduce_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.chunk_router_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            s = msg.response.search.topk_results.add()
            s.score = 0.1
            s.score_explained = '1-c1'
            s.chunk.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.score_explained = '1-c2'
            s.chunk.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.score_explained = '1-c3'
            s.chunk.doc_id = 1

            msg.envelope.num_part.extend([1, 2])
            c1.send_message(msg)

            msg.response.search.ClearField('topk_results')

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.score_explained = '2-c1'
            s.chunk.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.score_explained = '2-c2'
            s.chunk.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.score_explained = '2-c3'
            s.chunk.doc_id = 3
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            self.assertEqual(len(r.response.search.topk_results), 3)
            self.assertGreaterEqual(r.response.search.topk_results[0].score, r.response.search.topk_results[-1].score)
            print(r.response.search.topk_results)
            self.assertEqual(r.response.search.topk_results[0].score_explained, '1-c1\n1-c3\n2-c1\n')
            self.assertEqual(r.response.search.topk_results[1].score_explained, '1-c2\n2-c2\n')
            self.assertEqual(r.response.search.topk_results[2].score_explained, '2-c3\n')

            self.assertAlmostEqual(r.response.search.topk_results[0].score, 0.6)
            self.assertAlmostEqual(r.response.search.topk_results[1].score, 0.4)
            self.assertAlmostEqual(r.response.search.topk_results[2].score, 0.3)

    def test_doc_reduce_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.doc_router_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()

            # shard1 only has d1
            s = msg.response.search.topk_results.add()
            s.score = 0.1
            s.doc.doc_id = 1
            s.doc.raw_text = 'd1'

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.doc.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.doc.doc_id = 3

            msg.envelope.num_part.extend([1, 2])
            c1.send_message(msg)

            msg.response.search.ClearField('topk_results')

            # shard2 has d2 and d3
            s = msg.response.search.topk_results.add()
            s.score = 0.1
            s.doc.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.doc.doc_id = 2
            s.doc.raw_text = 'd2'

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.doc.doc_id = 3
            s.doc.raw_text = 'd3'

            msg.response.search.top_k = 5
            c1.send_message(msg)
            r = c1.recv_message()

            print(r.response.search.topk_results)
            self.assertSequenceEqual(r.envelope.num_part, [1])
            self.assertEqual(len(r.response.search.topk_results), 3)
            self.assertGreaterEqual(r.response.search.topk_results[0].score, r.response.search.topk_results[-1].score)

    def test_chunk_sum_reduce_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.chunk_sum_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            s = msg.response.search.topk_results.add()
            s.score = 0.6
            s.score_explained = '1-c1\n1-c3\n2-c1\n'
            s.doc.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score = 0.4
            s.score_explained = '1-c2\n2-c2\n'
            s.doc.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.score_explained = '2-c3\n'
            s.doc.doc_id = 3

            msg.envelope.num_part.extend([1, 2])
            c1.send_message(msg)

            msg.response.search.ClearField('topk_results')

            s = msg.response.search.topk_results.add()
            s.score = 0.5
            s.score_explained = '2-c1\n1-c2\n1-c1\n'
            s.doc.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.score_explained = '1-c3\n2-c2\n'
            s.doc.doc_id = 3

            s = msg.response.search.topk_results.add()
            s.score = 0.1
            s.score_explained = '2-c3\n'
            s.doc.doc_id = 1
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            self.assertEqual(len(r.response.search.topk_results), 3)
            self.assertGreaterEqual(r.response.search.topk_results[0].score, r.response.search.topk_results[-1].score)
            print(r.response.search.topk_results)
            self.assertEqual(r.response.search.topk_results[0].score_explained, '1-c2\n2-c2\n\n2-c1\n1-c2\n1-c1\n\n')
            self.assertEqual(r.response.search.topk_results[1].score_explained, '1-c1\n1-c3\n2-c1\n\n2-c3\n\n')
            self.assertEqual(r.response.search.topk_results[2].score_explained, '2-c3\n\n1-c3\n2-c2\n\n')

            self.assertAlmostEqual(r.response.search.topk_results[0].score, 0.9)
            self.assertAlmostEqual(r.response.search.topk_results[1].score, 0.7)
            self.assertAlmostEqual(r.response.search.topk_results[2].score, 0.6)

    def test_doc_sum_reduce_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.doc_sum_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()

            s = msg.response.search.topk_results.add()
            s.score = 0.4
            s.doc.doc_id = 1
            s.doc.raw_text = 'd3'
            s.score_explained = '1-d3\n'

            s = msg.response.search.topk_results.add()
            s.score = 0.3
            s.doc.doc_id = 2
            s.doc.raw_text = 'd2'
            s.score_explained = '1-d2\n'

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.doc.doc_id = 3
            s.doc.raw_text = 'd1'
            s.score_explained = '1-d3\n'

            msg.envelope.num_part.extend([1, 2])
            c1.send_message(msg)

            msg.response.search.ClearField('topk_results')

            s = msg.response.search.topk_results.add()
            s.score = 0.5
            s.doc.doc_id = 1
            s.doc.raw_text = 'd2'
            s.score_explained = '2-d2\n'

            s = msg.response.search.topk_results.add()
            s.score = 0.2
            s.doc.doc_id = 2
            s.doc.raw_text = 'd1'
            s.score_explained = '2-d1\n'

            s = msg.response.search.topk_results.add()
            s.score = 0.1
            s.doc.doc_id = 3
            s.doc.raw_text = 'd3'
            s.score_explained = '2-d3\n'

            msg.response.search.top_k = 5
            c1.send_message(msg)
            r = c1.recv_message()

            print(r.response.search.topk_results)
            self.assertSequenceEqual(r.envelope.num_part, [1])
            self.assertEqual(len(r.response.search.topk_results), 3)
            self.assertGreaterEqual(r.response.search.topk_results[0].score, r.response.search.topk_results[-1].score)

    def test_concat_router(self):
        args = set_router_service_parser().parse_args([
            '--yaml_path', self.concat_router_yaml,
            '--socket_out', str(SocketType.PUSH_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            msg.request.search.query.chunk_embeddings.CopyFrom(array2blob(np.random.random([5, 2])))
            msg.envelope.num_part.extend([1, 3])
            c1.send_message(msg)
            c1.send_message(msg)
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            print(r.envelope.routes)
            self.assertEqual(r.request.search.query.chunk_embeddings.shape, [5, 6])

            for j in range(1, 4):
                d = msg.request.index.docs.add()
                d.chunk_embeddings.CopyFrom(array2blob(np.random.random([5, 2 * j])))

            c1.send_message(msg)
            c1.send_message(msg)
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            for j in range(1, 4):
                self.assertEqual(r.request.index.docs[j - 1].chunk_embeddings.shape, [5, 6 * j])

    def test_multimap_multireduce(self):
        # p1 ->
        #      p21 ->
        #              r311
        #              r312
        #                   ->  r41
        #                             -> r5
        #      p22 ->
        #              r321
        #              r322
        #                   -> r42
        #                             -> r5
        #                                       -> client
        p1 = set_router_service_parser().parse_args([
            '--yaml_path', self.publish_router_yaml,
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUB_BIND),
        ])
        r5 = set_router_service_parser().parse_args([
            '--yaml_path', self.reduce_router_yaml,
            '--socket_in', str(SocketType.PULL_BIND),
            '--socket_out', str(SocketType.PUSH_CONNECT),
        ])
        r41 = set_router_service_parser().parse_args([
            '--yaml_path', self.reduce_router_yaml,
            '--socket_in', str(SocketType.PULL_BIND),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_out', str(r5.port_in)
        ])
        r42 = set_router_service_parser().parse_args([
            '--yaml_path', self.reduce_router_yaml,
            '--socket_in', str(SocketType.PULL_BIND),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_out', str(r5.port_in)
        ])
        p21 = set_router_service_parser().parse_args([
            '--yaml_path', self.publish_router_yaml,
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUB_BIND),
            '--port_in', str(p1.port_out)
        ])
        p22 = set_router_service_parser().parse_args([
            '--yaml_path', self.publish_router_yaml,
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUB_BIND),
            '--port_in', str(p1.port_out)
        ])
        r311 = set_router_service_parser().parse_args([
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_in', str(p21.port_out),
            '--port_out', str(r41.port_in),
            '--yaml_path', 'BaseRouter'
        ])
        r312 = set_router_service_parser().parse_args([
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_in', str(p21.port_out),
            '--port_out', str(r41.port_in),
            '--yaml_path', 'BaseRouter'
        ])
        r321 = set_router_service_parser().parse_args([
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_in', str(p22.port_out),
            '--port_out', str(r42.port_in),
            '--yaml_path', 'BaseRouter'
        ])
        r322 = set_router_service_parser().parse_args([
            '--socket_in', str(SocketType.SUB_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--port_in', str(p22.port_out),
            '--port_out', str(r42.port_in),
            '--yaml_path', 'BaseRouter'
        ])

        c_args = _set_client_parser().parse_args([
            '--port_in', str(r5.port_out),
            '--port_out', str(p1.port_in),
            '--socket_in', str(SocketType.PULL_BIND),
            '--socket_out', str(SocketType.PUSH_BIND),
        ])
        with RouterService(p1), RouterService(r5), \
             RouterService(p21), RouterService(p22), \
             RouterService(r311), RouterService(r312), RouterService(r321), RouterService(r322), \
             RouterService(r41), RouterService(r42), \
             ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            msg.envelope.num_part.append(1)
            c1.send_message(msg)
            r = c1.recv_message()
            self.assertSequenceEqual(r.envelope.num_part, [1])
            print(r.envelope.routes)
